#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict

import jsonpickle


class JsonObject:
    """Model to json - base class."""

    def json_str(self) -> str:
        """Return object as json string."""
        return jsonpickle.encode(self, make_refs=False, unpicklable=False)

    def json(self) -> Dict:
        """Return object as json."""
        return jsonpickle.decode(self.json_str())
